UNIT ModLoader;

INTERFACE
USES Objects, SongUnit;

PROCEDURE LoadModFileFormat(VAR Song  : tSong; VAR St  : tStream; VAR Header  : tSongHeader);

IMPLEMENTATION
USES SongElements, Heaps, AsciiZ;

{----------------------------------------------------------------------------}
{ Internal definitions. Format of the files.                                 }
{____________________________________________________________________________}

TYPE tModFileMagic = ARRAY[0..3] of Char;

CONST Mod31MagicM_K_  : tModFileMagic = ( 'M', '.', 'K', '.' );
      Mod31MagicFLT4  : tModFileMagic = ( 'F', 'L', 'T', '4' );
      Mod31Magic6CHN  : tModFileMagic = ( '6', 'C', 'H', 'N' );
      Mod31Magic8CHN  : tModFileMagic = ( '8', 'C', 'H', 'N' );

{ Instrument in a MOD file. 30 Bytes. }
TYPE tModFileInstrument = RECORD
       Name       : ARRAY [1..22] OF Char; { AsciiZ string, name of the instrument. }
       Len        : Word;                  { Length of the sample DIV 2.            }
       FineTune,                           { Fine tuning value.                     }
       Vol        : Byte;                  { Default volume.                        }
       LoopStart,                          { Offset of the loop DIV 2.              }
       LoopLen    : Word;                  { Length of the loop DIV 2.              }
     END;

{ Note in the file. 4 Bytes. }

     pModFileNote = ^tModFileNote;
     tModFileNote = RECORD
       CASE Integer OF
         1 : (L              : LongInt);
         2 : (W1, W2         : Word);
         3 : (B1, B2, B3, B4 : Byte);
     END;

     pModFilePattern = ^tModFilePattern;
     tModFilePattern = RECORD
       CASE Byte OF
         4 : (Patt4 : ARRAY[0..63] of ARRAY[1..4] of tModFileNote);
         5 : (Patt5 : ARRAY[0..63] of ARRAY[1..5] of tModFileNote);
         6 : (Patt6 : ARRAY[0..63] of ARRAY[1..6] of tModFileNote);
         7 : (Patt7 : ARRAY[0..63] of ARRAY[1..7] of tModFileNote);
         8 : (Patt8 : ARRAY[0..63] of ARRAY[1..8] of tModFileNote);
       END;

{ 15 samples module header format. 600 Bytes. }

       pModFile15 = ^tModFile15;
       tModFile15 = RECORD
         Name        : ARRAY[1..20] OF Char;               { AsciiZ song name.                   }
         Samples     : ARRAY[1..15] OF tModFileInstrument; { Instruments.                        }
         SongLen     : Byte;                               { Length of the sequency of the song. }
         SongRep     : Byte;                               { Song loop start position.           }
         PatternList : ARRAY[0..127] OF Byte;              { Pattern sequencies.                 }
       END;

{ 31 samples module header format. 1084 Bytes. }

         pModFile31 = ^tModFile31;
         tModFile31 = RECORD
           Name        : ARRAY[1..20] OF Char;               { AsciiZ song name.                   }
           Samples     : ARRAY[1..31] OF tModFileInstrument; { Instruments.                        }
           SongLen     : Byte;                               { Length of the sequency of the song. }
           SongRep     : Byte;                               { Song loop start position.           }
           PatternList : ARRAY[0..127] OF Byte;              { Pattern sequencies.                 }
           Magic       : tModFileMagic;                      { Magic number ("M.K.", "FLT4", etc.) }
         END;

PROCEDURE ProcessPatterns(VAR Song : tSong; VAR St : tStream; Num : Word);
VAR Patt       : tModFilePattern;
    FullTrack  : tFullTrack;
    Pattern    : PPattern;
    Track      : PTrack;
    I, J, N, T : Word;
    L          : LongInt;
BEGIN
  T := 1;
  For N := 1 to Num do
  BEGIN
    Pattern := Song.GetPattern(N);
    If Pattern = NIL then
    BEGIN
      Song.Status := msOutOfMemory;
      Exit;
    END;

    WITH Pattern^.Patt^ do
    BEGIN
      NNotes   := 64;
      NChans   := Song.NumChannels;
      Tempo    := 0;
      BPM      := 0;
    END;

    L := St.GetPos;
    St.Read(Patt, 64*4*Song.NumChannels);

    If St.Status <> stOk then
    BEGIN
      Song.Status := msFileTooShort;
      Exit;
    END;

    CASE Song.NumChannels OF
      4 : For I := 63 downto 0 do
            For J := Song.NumChannels downto 1 do
              Patt.Patt8[I][J] := Patt.Patt4[I][J];
      5 : For I := 63 downto 0 do
            For J := Song.NumChannels downto 1 do
              Patt.Patt8[I][J] := Patt.Patt5[I][J];
      6 : For I := 63 downto 0 do
            For J := Song.NumChannels downto 1 do
              Patt.Patt8[I][J] := Patt.Patt6[I][J];
      7 : For I := 63 downto 0 do
            For J := Song.NumChannels downto 1 do
              Patt.Patt8[I][J] := Patt.Patt7[I][J];
    END;

    For J := 1 to Song.NumChannels do
    BEGIN
      FillChar(FullTrack, SizeOf(FullTrack), 0);

      For I := 0 to 63 do
        WITH FullTrack[I], Patt.Patt8[I][J] Do
        BEGIN
          Command := tModCommand((B3 AND $F) + 1);
          If Command = mcExtended then
          BEGIN
            Parameter := B4 AND $F;
            Command   := tModCommand(($11 + (B4 SHR 4)));
          END else
          If (Command = mcArpeggio) AND (B4 = 0) then
          BEGIN
            Parameter := 0;
            Command   := mcNone;
          END else Parameter := B4;

          Period     := B2 + (Word(B1 AND $7) SHL 8);
          Instrument := (B3 SHR 4) + (B1 AND 16);

          If ((Command = mcEndPattern) or (Command = mcJumpPattern)) and (Pattern^.Patt^.NNotes>I+1)
            then Pattern^.Patt^.NNotes := I + 1;

          If (Command = mcSetVolume) and (Parameter > $40) then Parameter := $40;
          If (Command = mcJumpPattern) then Parameter := (Parameter AND $0F) + (Parameter SHR 4)*10+1;
          IF (Command = mcEndPattern) then Parameter := (Parameter AND 63) + 1;
        END;

      Track := Song.GetTrack(T);
      If Track = NIL then
      BEGIN
        Song.Status := msOutOfMemory;
        Exit;
      END;

      Track^.SetFullTrack(FullTrack);
      Pattern^.Patt^.Channels[J] := T;

      Inc(T);
    END;
  END;
END;

PROCEDURE ProcessInstruments(VAR Song : tSong; VAR St : tStream; Mod31 : tModFile31);
VAR Instrument : tInstrumentRec;
    Instr      : pInstrument;
    I          : Word;
BEGIN
  For I := 1 to 31 do
    WITH Instrument DO
    BEGIN
      FillChar(Instrument, SizeOf(Instrument), 0);
      Instr := Song.GetInstrument(I);
      If Instr = NIL then
      BEGIN
        Song.Status := msOutOfMemory;
        Exit;
      END;

      Instr^.SetName(StrASCIIZ(Mod31.Samples[I].Name, 22));
      Len  := LongInt(Swap(Mod31.Samples[I].Len)) SHL 1;

      If Len>St.GetSize - St.GetPos then
      BEGIN
        Len := St.GetSize - St.GetPos;
        Song.Status := msFileTooShort;
      END;

      If Len > 0 then
      BEGIN
        Reps := LongInt(Swap(Mod31.Samples[I].LoopStart)) SHL 1;
        Repl := LongInt(Swap(Mod31.Samples[I].LoopLen)  ) SHL 1;
        Vol  :=              Mod31.Samples[I].Vol;

        If Repl        > Len then Repl := Len;
        If Reps + Repl > Len then Reps := Len - Repl;

        If Mod31.Samples[I].Vol > $40 then Mod31.Samples[I].Vol := $40;
        If Vol>$40 then Vol := $40;

        If Len <= MaxSample then
        BEGIN
          FullHeap.HGetMem(Pointer(Data), Len);
          If Data = NIL then
          BEGIN
            Song.Status := msOutOfMemory;
            Exit;
          END;

          St.Read(Data^, Len);

          If St.Status <> stOk then
          BEGIN
            Song.Status := msFileDamaged;
            Exit;
          END;

{
          For W := 0 to Len - 1 do
            If Instruments[I].Data^[W] = -128 then Instruments[I].Data^[W] := -127;
}
        END else
        BEGIN
          FullHeap.HGetMem(Pointer(Data), MaxSample);
          FullHeap.HGetMem(Pointer(Xtra), Len-MaxSample);

          If (Data = NIL) or (Xtra = NIL) then
          BEGIN
            Song.Status := msOutOfMemory;
            Exit;
          END;

          St.Read(Data^, MaxSample);
          St.Read(Xtra^, Len-MaxSample);

          If St.Status <> 0 THEN
          BEGIN
            Song.Status := msFileDamaged;
            Exit;
          END;
        END;

        Instr^.Change(@Instrument);
      END;
    END;
END;


PROCEDURE LoadMod(VAR Song : tSong; VAR St : tStream; VAR Mod31 : tModFile31);
VAR J, K, I, W       : Word;
    IsMod31          : Boolean;
    NumberOfPatterns : Word;

BEGIN
  { Initial checkings to see if it's a real MOD. }
  Song.Status := msFileDamaged;

  For I := 0 to 127 do
    If Mod31.PatternList[I] > 63 then Exit;

  For I := 1 to 20 do
    If (Mod31.Name[I] < ' ') and (Mod31.Name[I] <> #0) then Exit;

  For J := 1 to 31 do
    For I := 1 to 22 do
      If (Mod31.Samples[j].Name[I] < ' ') and (Mod31.Samples[j].Name[I] <> #0) then Exit;

  If (Mod31.SongLen > 128) or (Mod31.SongRep > 128) then Exit;

  { Processing of the header }
  Song.Status := msOK;
  Song.Name := FullHeap.HNewStr(StrASCIIZ(Mod31.Name, 20));

  Song.InitialTempo := 6;
  Song.InitialBPM   := 125;
  Song.Volume       := 255;

  For I := 0 to 127 do
    Inc(Mod31.PatternList[I]);
  Move(Mod31.PatternList, Song.PatternSequence^, Mod31.SongLen);
  Song.SequenceLength   := Mod31.SongLen;
  Song.SequenceRepStart := Mod31.SongRep + 1;

  NumberOfPatterns := 0;
  For I := 0 to 127 do
    If NumberOfPatterns < Mod31.PatternList[I] then NumberOfPatterns := Mod31.PatternList[I];

  { Processing of the patterns (the partiture) }
  ProcessPatterns(Song, St, NumberOfPatterns);
  If Song.Status > msOk then Exit;

  { Processing of the instruments }
  ProcessInstruments(Song, St, Mod31);
  If Song.Status > msFileTooShort then Exit;
END;

PROCEDURE LoadMod15(VAR Song : tSong; VAR St : tStream; VAR Header : tSongHeader);
VAR I     : Word;
    Mod31 : tModFile31 ABSOLUTE Header;
    Mod15 : tModFile15 ABSOLUTE Header;
BEGIN
  Move(Mod15.SongLen, Mod31.SongLen, 130);
  For I := 16 to 31 do
    FillChar(Mod31.Samples[I], SizeOf(Mod31.Samples[I]), 0);

  St.Seek(St.GetPos - SizeOf(TModFile31) + SizeOf(TModFile15));

  LoadMod(Song, St, Mod31);
END;

PROCEDURE LoadModFileFormat(VAR Song : tSong; VAR St : tStream; VAR Header : tSongHeader);
VAR Mod31 : tModFile31 ABSOLUTE Header;
BEGIN
  St.Seek(St.GetPos + SizeOf(TModFile31));

  If (Mod31.Magic = Mod31MagicM_K_) then
  BEGIN
    If Song.FileExt = '.WOW' then
    BEGIN
      Song.NumChannels := 8;
      Song.FileFormat  := mffWow8;
    END else
    BEGIN
      Song.NumChannels := 4;
      Song.FileFormat  := mffMod31M_K_;
    END;

    LoadMod(Song, St, Mod31);
  END else
  If (Mod31.Magic = Mod31MagicFLT4) then
  BEGIN
    Song.NumChannels := 4;
    Song.FileFormat  := mffMod31FLT4;
    LoadMod(Song, St, Mod31);
  END else
  If (Mod31.Magic = Mod31Magic6CHN) then
  BEGIN
    Song.NumChannels := 6;
    Song.FileFormat  := mffFastTracker;
    LoadMod(Song, St, Mod31);
  END else
  If (Mod31.Magic = Mod31Magic8CHN) then
  BEGIN
    Song.NumChannels := 8;
    Song.FileFormat  := mffFastTracker;
    LoadMod(Song, St, Mod31);
  END else
  BEGIN
    Song.NumChannels := 4;
    Song.FileFormat  := mffMod15;
    LoadMod15(Song, St, Header);
  END
END;

END.
