UNIT PlayMod;

INTERFACE
USES SongUnit, SongUtils, SongElements, ModCommands, SoundDevices, Filters, Kbd;

{ Definitions. }
TYPE tTickProc = PROCEDURE(VAR Song  : tSong; Note : Boolean); { Procedure to execute every tick. }
     tVolumes  = ARRAY[1..MaxChannels] of Byte;               { Volume set (all channels). }

{ General definitions about the way of playing the music. }
{ Music player configuration.                             }

CONST PlayingSong    : pSong         = Nil;
      LoopMod        : Boolean       = True;   { True if music can be played forever.                  }
      ForceLoopMod   : Boolean       = False;  { True if music must be played forever.                 }
      CanFallBack    : Boolean       = True;   { True if fall-back is allowed.                         }
      FilterOn        : tFilterMethod = fmNone; { Initial value of the ON  filter.                      }
      FilterOff       : tFilterMethod = fmNone; { Initial value of the OFF filter.                      }
      FilterIsOn     : Boolean       = False;  { Initial position of the filter (False = OFF).         }
      TicksPerSecond : Word          = 50;     { Number of ticks per second, 50 = Europe, 60 = USA.    }
      MaxOutputFreq  : Word          = 44000;  { Maximum frequency of the output sound.                }
                                               { Less means less memory for buffers.                   }

VAR SplBuf : ARRAY[1..MaxChannels] of Word;

{ Exported variables. }
CONST Playing          : Boolean = False;      { (Read only) True if the music is sounding right now.     }
      ModTickProcValid : Boolean = False;      { True if the module tick procedure has been initialised.  }

VAR ActualHz      : Word;                   { Desired freq. of the sound.                              }
    NoteHz        : Word;                   { Freq. to be used in the current tick.                    }
    UserVols       : tVolumes;               { Channel volumes.                                         }
    Permisos      : ARRAY[1..MaxChannels] OF Boolean; { Permissions for playing the channels.          }
    TickCount     : Word;                   { Ticks counter. Increments each tick.                     }
    ModTickProc    : tTickProc;              { Tick procedure. A procedure to be executed every tick.   }
    MyCanFallBack : Boolean;                { Actual permission to fall-back.                          }
    FilterVal      : tFilterMethod;          { Method of the filter to be used.                         }

{ Definition of the local stack. }
CONST PlayModStackSize = 500;               { Size of the stack. }

VAR PlayModStack : ARRAY[1..PlayModStackSize] of Byte;

{ Definitions concerning a note. Buffer of the last N notes. }
TYPE
  pPlayingNote = ^tPlayingNote;
  tPlayingNote = RECORD
    EoMod       : Boolean;                 { True if it is the note following the last.                }
    Tempo       : Byte;                    { Number of ticks the note will last.                       }
    NotePlaying : Byte;                    { Index of the note inside the pattern.                     }
    SeqPlaying  : Byte;                    { Sequence number of the pattern to which the note belongs. }
    Volume       : tVolumes;                { Volumes of the channels.                                  }
    Note        : ARRAY[1..MaxChannels] of tFullNote; { Notes of the channels.                         }
    NMuestras   : Word;                    { Number of samples processed for each note.                }
    Fill        : Byte;                    { To make it a 32-Byte record.                              }
  END;

CONST NoteBuffSize = 1; { Number of note buffers. }

VAR NoteBuff : ARRAY[0..NoteBuffSize-1] OF tPlayingNote;

CONST NoteTl : Word = 0;
      NoteHd : Word = 0;

      NoteSound     : pPlayingNote = Nil;
      NoteProcessed : pPlayingNote = Nil;

VAR Canales : ARRAY[1..MaxChannels] of tCanal; { State of the channels. }

{----------------------------------------------------------------------------}
{ Definition of the buffers where the samples are placed.                    }
{____________________________________________________________________________}
CONST MaxSplPerTick : Word = 880; { Maximum samples in the buffer. Means maximum samples per tick. }
      NumBuffers           = 3;   { Number of buffers.                                             }

VAR BuffIdx,         { Tail of the buffer. }
    BuffGive : Word; { Head of the buffer. }

    Buffers       : ARRAY[1..NumBuffers] of tSampleBuffer;
    SizeOfABuffer : Word;

{----------------------------------------------------------------------------}
{ Exported procedures.                                                       }
{____________________________________________________________________________}

PROCEDURE PlayStart(VAR Song : tSong);
PROCEDURE PlayStop;

PROCEDURE ChangeSamplingRate(Hz : Word);
PROCEDURE ProcessTickEntry;
PROCEDURE FillWithSamples(VAR Buff; Size : Word);

IMPLEMENTATION
USES Dos, Heaps, Debugging;

{----------------------------------------------------------------------------}
{ General definitions of the module player. They define its actual state.    }
{____________________________________________________________________________}
VAR DelaySamples    : Boolean;     { True means it couldn't fill the samples buffer.                }
    MuestrasPerTick : Word;        { Number of samples that there are in a tick at the actual freq. }

{----------------------------------------------------------------------------}
{ Raw channel definitions.                                                   }
{____________________________________________________________________________}
TYPE pModRawChan = ^tModRawChan;
     tModRawChan = RECORD
       Flags      : Byte;         { Channel flags (see below).                 }

       SplPosFrac : Byte;         { Position fraction.                         }
       SplPosInt  : Word;         { Position offset.                           }
       SplPosSeg  : Word;         { Position segment.                          }

       SplOfs     : Word;         { Actual sample part offset.                 }
       SplSeg     : Word;         { Actual sample part segment.                }
       SplLimit   : Word;         { Actual sample part size.                   }

       SplOfs1    : Word;         { First  sample part offset.                 }
       SplSeg1    : Word;         { First  sample part segment.                }
       SplLimit1  : Word;         { First  sample part size.                   }

       SplOfs2    : Word;         { Second sample part offset.                 }
       SplSeg2    : Word;         { Second sample part segment.                }
       SplLimit2  : Word;         { Second sample part size.                   }

       StepFrac   : Byte;         { Sample incement fraction.                  }
       StepInt    : Word;         { Sample incement integer.                   }

       Volume     : Byte;         { Volume to be used.                         }

       LoopEnd    : Word;         { Offset of the end of the loop in its part. }
       LoopLen    : Word;         { Size of the loop in its part.              }
     END;

{ TModRawChan.Flags }
CONST rcfLongSample     = $01; { Set if it's a long (more than 65520 Bytes) sample.     }
      rcfActiveChannel  = $02; { Set if the channel is activated (permission to sound). }
      rcfDoesLoop       = $04; { Set of the sample has a loop.                          }
      rcfPlaying2nd     = $08; { Set if playing the 2nd part of the long loop.          }
      rcfLongLoopLen    = $10; { Loop size goes from the 2nd part to the 1st.           }
      rcfLongLoopEnd    = $20; { Loop ends in the 2nd part.                             }
      rcfSampleFinished = $40; { Set if the sample has already finished.                }

{ Raw channels }
VAR RawChannels : ARRAY[1..MaxChannels] OF tModRawChan;

{----------------------------------------------------------------------------}
{ Basic, fast assembler routines.                                            }
{____________________________________________________________________________}
PROCEDURE UnCanal; ASSEMBLER;
ASM
  TEST    [tModRawChan(DS:BX).Flags],rcfActiveChannel
  JZ      @@Desactivado

  TEST    [TModRawChan(DS:BX).Flags],rcfSampleFinished
  JNZ     @@Desactivado

  TEST    Byte PTR [TModRawChan(DS:BX).Volume],$FF
  JZ      @@Desactivado

  PUSH    BX

  TEST    [TModRawChan(DS:BX).Flags],rcfDoesLoop
  JZ      @@NoDoesLoop

  MOV     AX,[TModRawChan(DS:BX).LoopEnd]
  MOV     Word PTR [CS:@@dlData2-2],AX

  MOV     AX,[TModRawChan(DS:BX).LoopLen]
  MOV     Word PTR [CS:@@dlData3-2],AX

  MOV     DL,Byte PTR [TModRawChan(DS:BX).Volume]
  MOV     AL,[TModRawChan(DS:BX).StepFrac]
  MOV     BP,[TModRawChan(DS:BX).StepInt]

  MOV     AH,[TModRawChan(DS:BX).SplPosFrac]

  LDS     SI,DWord PTR [TModRawChan(DS:BX).SplPosInt]

  MOV     BX,AX

@@dlLoop:
  MOV     AL,[SI]
  IMUL    DL
  MOV     [ES:DI],AX
  ADD     DI,MaxChannels*2

@@dlData1:
  ADD     BH,BL
  ADC     SI,BP
  JC      @@dlSplLoop


@@dlChkLoop:
  CMP     SI,$1234

@@dlData2:
  JB      @@dlNoLoop

@@dlSplLoop:
  SUB     SI,$1234

@@dlData3:


@@dlNoLoop:
  LOOP    @@dlLoop

  JMP     @@Finish

@@NoDoesLoop:
  MOV     AX,[TModRawChan(DS:BX).SplLimit]
  MOV     Word PTR [CS:@@nlData2-2],AX

  MOV     DL,Byte PTR [TModRawChan(DS:BX).Volume]
  MOV     AL,[TModRawChan(DS:BX).StepFrac]
  MOV     AH,[TModRawChan(DS:BX).SplPosFrac]

  MOV     BP,[TModRawChan(DS:BX).StepInt]

  LDS     SI,DWord PTR [TModRawChan(DS:BX).SplPosInt]

  MOV     BX,AX

@@nlLoop:
  MOV     AL,[SI]
  IMUL    DL
  MOV     [ES:DI],AX
  ADD     DI,MaxChannels*2

@@nlData1:
  ADD     BH,BL
  ADC     SI,BP
  JC      @@nlSeguroFin
  CMP     SI,$1234

@@nlData2:
  JNB     @@nlSeguroFin

@@nlNoLoop:
  LOOP    @@nlLoop
  JMP     @@Finish

@@nlSeguroFin:
  MOV     BX,SEG @Data
  MOV     DS,BX
  POP     BX
  OR      Byte PTR [TModRawChan(DS:BX).Flags],rcfSampleFinished
  DEC     CX
  JCXZ    @@Fin

@@Desactivado:
  XOR     AX,AX

@@Data2:
  MOV     [ES:DI],AX
  ADD     DI,MaxChannels*2

@@Data1:
  LOOP    @@Data2
  JMP     @@Fin

@@Finish:
  MOV     BP,SEG @Data
  MOV     DS,BP
  POP     BP
  MOV     [TModRawChan(DS:BP).SplPosInt],SI
  MOV     [TModRawChan(DS:BP).SplPosFrac],BH

@@Fin:
  MOV     AX,SEG @Data
  MOV     DS,AX
END;

PROCEDURE SetNewSample(VAR Raw : tModRawChan; Spl : pInstrumentRec);
CONST _Or : Byte    = 0;
       F  : Boolean = False;
BEGIN
  FillChar(Raw, SizeOf(Raw), 0);
  If Spl = Nil then Exit;

  ASM
    MOV     DI,Word PTR Raw
    LES     SI,Spl

    MOV     AX,Word PTR TInstrumentRec([ES:SI]).data
    MOV     TModRawChan([DI]).SplOfs1,AX
    MOV     AX,Word PTR TInstrumentRec([ES:SI+2]).data
    MOV     TModRawChan([DI]).SplSeg1,AX
    MOV     _or,rcfActiveChannel

    MOV     AX,Word PTR TInstrumentRec([ES:SI+1]).repl
    and     AX,AX
    JNZ     @@1
    MOV     AL,Byte PTR TInstrumentRec([ES:SI]).repl
    CMP     AL,4
    JNB     @@1
    MOV     f,1
    JMP     @@2

@@1:
    MOV     f,0
    OR      _or,rcfDoesLoop
@@2:
END;

  If Spl^.len > MaxSample then
  BEGIN
    ASM
      MOV     DI,Word PTR Raw
      LES     SI,Spl

      OR      _or,rcfLongSample
      MOV     TModRawChan([DI]).SplLimit1,MaxSample
    END;

    Raw.SplLimit2 := Spl^.len - MaxSample;
    Raw.SplOfs2   := OFS(Spl^.xtra^);
    Raw.SplSeg2   := SEG(Spl^.xtra^);

    If NOT f then
    BEGIN
      If (Spl^.reps > MaxSample) or (Spl^.reps+Spl^.repl <= MaxSample) then Raw.LoopLen := Spl^.repl else
      BEGIN
        Raw.LoopLen := Spl^.repl - MaxSample;
        Inc(_or, rcfLongLoopLen);
      END;
      If Spl^.reps+Spl^.repl <= MaxSample then Raw.LoopEnd := Spl^.reps + Spl^.Repl else
      BEGIN
        Raw.LoopEnd := Spl^.reps + Spl^.repl - MaxSample;
        Inc(_or, rcfLongLoopEnd);
      END;
    END;
  END else
  BEGIN
    ASM
      MOV     DI,Word PTR Raw
      LES     SI,Spl

      MOV     AX,Word PTR TInstrumentRec([ES:SI]).len
      MOV     TModRawChan([DI]).SplLimit1,AX

      MOV     AL,f
      and     AL,AL
      JNZ     @@1
      MOV     AX,Word PTR TInstrumentRec([ES:SI]).repl
      MOV     TModRawChan([DI]).LoopLen,AX
      ADD     AX,Word PTR TInstrumentRec([ES:SI]).reps
      MOV     TModRawChan([DI]).LoopEnd,AX
    @@1:
    END;
  END;

  ASM
    MOV     DI,Word PTR Raw
    MOV     TModRawChan([DI]).SplPosFrac,0
    MOV     AX,TModRawChan([DI]).SplOfs1
    MOV     TModRawChan([DI]).SplPosInt,AX
    MOV     TModRawChan([DI]).SplOfs,AX
    MOV     AX,TModRawChan([DI]).SplSeg1
    MOV     TModRawChan([DI]).SplPosSeg,AX
    MOV     TModRawChan([DI]).SplSeg,AX
    MOV     AX,TModRawChan([DI]).SplLimit1
    MOV     TModRawChan([DI]).SplLimit,AX
    MOV     AL,_or
    MOV     TModRawChan([DI]).Flags,AL
  END;
END;

PROCEDURE MyMove(VAR Src, Dest; Bytes : Word); ASSEMBLER;
ASM
  PUSH    DS

  LDS     SI,[Src]
  LES     DI,[Dest]
  MOV     CX,[Bytes]

  CLD

  AND     CX,CX
  JZ      @@Fin

  TEST    SI,1
  JZ      @@nobeg
  MOVSB
  DEC     CX
  JZ      @@Fin

@@nobeg:
  MOV     BX,CX
  SHR     CX,1
  REP     MOVSW
  MOV     CX,BX
  AND     CL,1
  JZ      @@Fin

  MOVSB

@@Fin:
  POP     DS
END;

PROCEDURE ProcessNewNote(VAR Song : tSong);
CONST I    : Word      = 0;
      J    : Word      = 0;
      N    : tFullNote = (Instrument:0);
      Can  : ^TCanal   = Nil;
      Patt : PPattern  = Nil;
BEGIN
  I := (NoteHd + 1) and (NoteBuffSize - 1);
  NoteProcessed := @NoteBuff[I];
  MyMove(NoteBuff[NoteHd], NoteProcessed^, SizeOf(NoteBuff[0]));
  NoteHd := i;
  WITH NoteProcessed^ DO
  BEGIN
    EoMod := NextNote = $FFFF;
    If EoMod then
      If MyLoopMod then
      BEGIN
        NextSeq := MyRepStart;
        If NextSeq < MyFirstPattern then NextSeq := MyFirstPattern;

        NextNote  := 1;
        EoMod     := False;
      END else
      BEGIN
        Playing := False;
        Exit;
      END;

    NotePlaying := NextNote;
    SeqPlaying  := NextSeq;
    Volume      := UserVols;

    Patt := Song.GetPatternSeq(SeqPlaying);

    If NextNote < Patt^.Patt^.NNotes then Inc(NextNote) else
    BEGIN
      Inc(NextSeq);
      If NextSeq > MySongLen
        then NextNote := $FFFF
        else NextNote := 1;
    END;

    If Song.GetPatternSequence(SeqPlaying) = 0 then
    BEGIN
      ModCommands.Tempo        := Song.InitialTempo;
      ModCommands.BPMIncrement := Song.InitialBPM;

      FillChar(Canales, SizeOf(Canales), 0);

      For I := 1 to MaxChannels do
        WITH Canales[I] DO
        BEGIN
          Note.Period     := 800;
          Note.Instrument := 1;
          Note.Command    := mcNone;
          Period          := 800;
        END;

      REPEAT
        Inc(SeqPlaying);
        If SeqPlaying > MySongLen
          then NextNote := $FFFF
          else NextNote := 2;
      UNTIL (NextNote = $FFFF) or (Song.GetPatternSequence(SeqPlaying) <> 0);

      NextSeq := SeqPlaying;
    END;

    If (NotePlaying = 1) and (Song.GetPatternTempo(SeqPlaying) <> 0)
      then ModCommands.Tempo := Song.GetPatternTempo(SeqPlaying);

    For J := 1 to Song.NumChannels do
    BEGIN
      Can := @Canales[J];

      Song.GetNote(SeqPlaying, NotePlaying, J, N);

      MyMove(N, Note[J], SizeOf(N));

      If ((N.Instrument <> 0) and (Can^.Note.Instrument <> N.Instrument)) or
         ((0 <> N.Period) and (N.Command <> mcNPortamento) and (N.Command <> mcT_VSlide)) then
      BEGIN
        If N.Instrument <> 0 then
        BEGIN
          Can^.Note.Instrument := N.Instrument;
          Can^.Instrument      := pInstrument(Song.GetInstrument(N.Instrument))^.Instr;
        END;

        SetNewSample(RawChannels[J], Can^.Instrument);
      END;

      If (N.Instrument <> 0) and (Can^.Instrument <> Nil)
        then Can^.Volume := Can^.Instrument^.Vol;

      If N.Volume <> 0 then Can^.Volume := N.Volume - 1;
      If Can^.Volume > 64 then Can^.Volume := 64;

      CommandStart(Song, Can^, N);

      NoteProcessed^.Tempo := ModCommands.Tempo;
    END;

    MuestrasPerTick := ActualHz DIV TicksPerSecond;

    If MuestrasPerTick > MaxSplPerTick then MuestrasPerTick := MaxSplPerTick;
    NMuestras := MuestrasPerTick * Tempo;
    NoteHz    := ActualHz;
  END;

  NoteTl    := NoteHd;
  NoteSound := NoteProcessed;
END;

PROCEDURE FillChannels(VAR Song : tSong);
CONST FirstTick : Boolean       = True;
      I         : Word          = 0;
      P         : ^tModRawChan  = Nil;
      Q         : Pointer       = Nil;
      Buf       : PSampleBuffer = Nil;
BEGIN
  Buf := @Buffers[BuffIdx];
  DelaySamples := Buf^.InUse;
  If DelaySamples then Exit;

  For I := 1 to Song.NumChannels do
  BEGIN
    P := @RawChannels[I];
    Q := Addr(Buf^.IData^[I-1]);
    ASM
      PUSH    BP
      PUSH    DI
      PUSH    SI
      PUSH    ES
      MOV     CX,MuestrasPerTick
      MOV     BX,Word PTR p
      LES     DI,q
      CALL    UnCanal
      POP     ES
      POP     SI
      POP     DI
      POP     BP
    END;

    SplBuf[I] := FilterChunkWord(Q^, MuestrasPerTick, MaxChannels, FilterVal, SplBuf[I]);
  END;

  Buf^.InUse    := True;
  Buf^.NSamples := MuestrasPerTick;
  Buf^.RateHz   := NoteHz;
  Buf^.DataType := dtInteger;
  Buf^.Channels := MaxChannels;

  Inc(BuffIdx);
  If BuffIdx > NumBuffers then BuffIdx := 1;
END;

PROCEDURE ProcessTick(VAR Song : tSong);
CONST SOTCanal = SizeOf(TCanal);
      incr        : Integer     = 0;
      OTempoCt    : Word        = 0;
      Can         : PCanal      = Nil;
      Raw         : pModRawChan = Nil;
      NoteHzFreq  : LongInt     = 0;
      i           : Word        = 0;
      j           : Word        = 0;
      step        : LongInt     = 0;
      FBCount     : Word        = 0;
      NumChannels : Byte        = 0;
LABEL Fin;

BEGIN
  If DelaySamples then
  BEGIN
    FillChannels(Song);
    If DelaySamples then Goto Fin;
  END;

  Inc(TickCount);

  OTempoCt := TempoCt;
  Inc(BPMCount, BPMIncrement);
  Inc(TempoCt, BPMCount DIV BPMDivider);
  If TempoCt <> OTempoCt then BPMCount := BPMCount MOD BPMDivider;

  IF TempoCt >= NoteProcessed^.Tempo then
  BEGIN
    ProcessNewNote(Song);
    If not Playing then Goto Fin;
    TempoCt := 0;
  END;

  If not MyCanFallBack then PleaseFallBack := 0;

  If PleaseFallBack > 0 then
  BEGIN
    PleaseFallBack := 0;
    I := ActualHz;
    WHILE (I = ActualHz) and (I <> ActiveDevice^.GetRealFreqProc(0)) DO
    BEGIN
      Dec(DesiredHz, 100);
      I := ActiveDevice^.GetRealFreqProc(DesiredHz);
    END;
    ChangeSamplingRate(DesiredHz);
  END;

  NumChannels := Song.NumChannels;

  If (TempoCt > 0) OR Song.FirstTick then
  ASM
    XOR    CH,CH
    MOV    CL,[NumChannels]

@@lp:
    PUSH   CX
    MOV    AL,CL
    DEC    AL
    MOV    BL,SOTCanal
    MUL    BL
    MOV    SI,OFFSET Canales
    ADD    SI,AX
    MOV    BL,TCanal([SI]).Note.Command
    ADD    BL,BL
    XOR    BH,BH
    CALL   DoTickCommand
    POP    CX
    LOOP   @@lp
  END;

  For I := 1 to Song.NumChannels do
  BEGIN
    Can := @Canales[I];
    Raw := @RawChannels[I];

    If not Permisos[I]
      then Raw^.Flags := Raw^.Flags and NOT rcfActiveChannel
      else Raw^.Flags := Raw^.Flags OR      rcfActiveChannel;

    Raw^.Volume := (Can^.Volume*Word(UserVols[I]) SHR 4) div ((Song.NumChannels + 1) and $FFFE);
    If Raw^.Volume >= $80 then Raw^.Volume := $7F;
    If Can^.Period = 0 then Can^.Period := 1;
    If NoteHz      = 0 then NoteHz      := 1;

    If Can^.Instrument <> Nil
      then Can^.RealPeriod := (LongInt(Can^.Period) * Can^.Instrument^.NAdj) DIV Can^.Instrument^.DAdj
      else Can^.RealPeriod := Can^.Period;

    ASM
      LES     DI,[Can]
      MOV     DX,TCanal([ES:DI]).RealPeriod
      MOV     AX,[NoteHz]
      MUL     DX
      MOV     Word PTR [NoteHzFreq],AX
      MOV     Word PTR [NoteHzFreq+2],DX
    END;

    Step := (65536 * 14000) DIV NoteHzFreq;

    Raw^.StepFrac := Lo(Step);
    Raw^.StepInt  := Step SHR 8;

    If FilterIsOn then FilterVal := FilterOn else FilterVal := FilterOff;
  END;

  FillChannels(Song);

Fin:
END;

PROCEDURE ProcessTickEntry;
CONST Semaphor : Byte = 0;
      _SS      : Word = 0;
      _SP      : Word = 0;
LABEL Fin1, Fin2;

BEGIN
  If not Playing then
  BEGIN
    TempoCT := 1;
    Goto Fin1;
  END;

  If Semaphor <> 0 then Goto Fin2;
  Inc(Semaphor);

  ASM
    MOV     [_SS],SS
    MOV     [_SP],SP
    MOV     AX,DS
    MOV     SS,AX
    MOV     SP,OFFSET PlayModStack + PlayModStackSize
  END;

  ProcessTick(PlayingSong^);

  ASM
    MOV     SS,[_SS]
    MOV     SP,[_SP]
  END;

  Dec(Semaphor);

Fin1:
  If ModTickProcValid then ModTickProc(PlayingSong^, TempoCt = 0);

Fin2:
END;

FUNCTION IdleGiver : pSampleBuffer; FAR;
BEGIN
  IdleGiver := Nil;
END;

FUNCTION BufferGiver : pSampleBuffer; FAR;
BEGIN
  BufferGiver := Nil;
  IF NOT Buffers[BuffGive].InUse then Exit;
  BufferGiver := @Buffers[BuffGive];
  Inc(BuffGive);
  If BuffGive > NumBuffers then BuffGive := 1;
END;

PROCEDURE FillWithSamples(VAR Buff; Size: Word);
CONST mBuff : pIntBuff = Nil;
BEGIN
  mBuff := Buffers[BuffGive].IData;

  ASM
    PUSH    DS

    XOR     SI,SI

    MOV     CX,[Size]
    MOV     AX,[MuestrasPerTick]
    AND     AX,AX
    JZ      @@bien
    CMP     AX,CX
    JNC     @@bien

    SUB     CX,AX
    MOV     SI,CX
    MOV     CX,AX

@@bien:
    CLD
    MOV     DX,16
    LDS     BX,[mBuff]
    LES     DI,[Buff]

@@lp:
    MOV     AX,[BX]
    ADD     AX,[BX+6]
    ADD     AX,[BX+8]
    ADD     AX,[BX+14]
    ADD     AX,[BX+16]
    ADD     AX,[BX+22]
    ADD     AX,[BX+24]
    ADD     AX,[BX+30]

    MOV     DX,[BX+2]
    ADD     DX,[BX+4]
    ADD     DX,[BX+10]
    ADD     DX,[BX+12]
    ADD     DX,[BX+18]
    ADD     DX,[BX+20]
    ADD     DX,[BX+26]
    ADD     DX,[BX+28]

    ADD     AX,DX
    JNO     @@nooverf
    JS      @@posit
    MOV     AX,-32768
    JMP     @@nooverf

@@posit:
    MOV     AX,32767

@@nooverf:
    ADD     BX,MaxChannels*2
    STOSW
    LOOP    @@lp
    AND     SI,SI
    JZ      @@Fin

    MOV     CX,SI
    XOR     AX,AX
    REP     STOSW

@@Fin:
    POP     DS
  END;
END;

PROCEDURE PlayStart(VAR Song : tSong);
VAR I, J : Word;
BEGIN
  ASM CLI END;
  PlayingSong := @Song;

  MyFirstPattern := FirstPattern;
  MyRepStart     := RepStart;
  MySongLen      := SongLen;

  If MySongLen = 0 then MySongLen := Song.SequenceLength;
  If MyFirstPattern = 0
    then NextSeq := 1
    else NextSeq := MyFirstPattern;

  If NextSeq > MySongLen then
  BEGIN
    ASM STI END;
    Exit;
  END;

  If (MyRepStart =  0) AND
     (Song.SequenceRepStart <= MySongLen) and
     (Song.SequenceRepStart <> 0) then
      MyRepStart    := Song.SequenceRepStart;

  MyLoopMod := (True and (MyRepStart <> 0)) OR ForceLoopMod;
  TempoCt         := 254;
  Tempo           := Song.InitialTempo;
  BPMIncrement    := Song.InitialBPM;
  TickCount       := 0;
  NextNote        := 1;
  DelaySamples    := False;
  MuestrasPerTick := 1;
  MaxSplPerTick   := MaxOutputFreq DIV TicksPerSecond;

  If MyRepStart < NextSeq then MyRepStart := NextSeq;

  WITH NoteBuff[0] DO
  BEGIN
    EoMod       := False;
    Tempo       := 6;
    NotePlaying := 0;
    SeqPlaying  := 0;
    Volume      := UserVols;
    NMuestras   := 0;
  END;

  NoteHd        := 0;
  NoteTl        := 0;
  NoteSound     := @NoteBuff[0];
  NoteProcessed := @NoteBuff[0];

  FillChar(Canales, SizeOf(Canales), 0);

  For I := 1 to MaxChannels do
    WITH Canales[I] DO
    BEGIN
      Note.Period     := 800;
      Note.Instrument := 1;
      Note.Command    := mcNone;
      Period          := 800;
    END;

  SizeOfABuffer := MaxSplPerTick*MaxChannels*2;
  FillChar(Buffers, SizeOf(Buffers),  0);
  For I := 1 to NumBuffers do
  BEGIN
    FullHeap.HGetMem(Pointer(Buffers[I].IData), SizeOfABuffer);
    If Buffers[I].IData = Nil then
    BEGIN
      Song.Status := msOutOfMemory;
      PlayStop;
      ASM STI END;
      Exit;
    END;
    FillChar(Buffers[I].IData^, SizeOfABuffer,  0);
  END;
  BuffIdx  := 1;
  BuffGive := 1;

  FillChar(RawChannels, SizeOf(RawChannels), 0);
  ChangeSamplingRate(DesiredHz);

  ASM STI END;

  SetBufferAsker(IdleGiver);
  MyCanFallBack  := False;
  Playing        := True;

  For I := 1 to NumBuffers do
    ProcessTickEntry;

  StartSampling;
  SetBufferAsker(BufferGiver);
  WHILE DeviceIdling DO;

  PleaseFallBack := 0;
  MyCanFallBack  := CanFallBack;
END;

PROCEDURE ChangeSamplingRate(Hz : Word);
VAR MyHz : Word;
LABEL Otra;
BEGIN
Otra:
  DesiredHz := Hz;
  MyHz      := ActiveDevice^.GetRealFreqProc(Hz);

  If MyHz >  MaxSplPerTick * TicksPerSecond then
  BEGIN
    Dec(Hz, 100);
    Goto Otra;
  END;

  If MyHz < 1000 then
  BEGIN
    Inc(Hz, 100);
    Goto Otra;
  END;

  If MyHz <> ActualHz then
  BEGIN
    ActualHz := MyHz;
    SetPeriodicProc(ProcessTickEntry, TicksPerSecond * 3);
  END;
END;

PROCEDURE PlayStop;
VAR I : Word;
BEGIN
  Playing := False;
  SetBufferAsker(IdleGiver);

  WHILE (NOT DeviceIdling) and (NOT KbdKeyPressed) DO;

  For I := 1 to NumBuffers do
    FullHeap.HFreeMem(Pointer(Buffers[I].IData), SizeOfABuffer);
END;

BEGIN
  Playing  := False;
  LoopMod  := False;
  ActualHz := 0;

  If FilterIsOn then FilterVal := FilterOn else FilterVal := FilterOff;

  FillChar(UserVols, SizeOf(UserVols), 255);
  FillChar(Permisos, SizeOf(Permisos), True);
  FillChar(SplBuf, SizeOf(SplBuf), 0);
END.
