UNIT SongElements;

INTERFACE
USES Objects;

{----------------------------------------------------------------------------}
{ Definitions for handling the format of individual notes.                   }
{ Notes are composed of four fields:                                         }
{                                                                            }
{   Period:     A number in the range 0..2047 which states the period of     }
{               the note in units of 1/3584000 per sample. (this is a        }
{               somewhat empyric number. If anyone knows the exact Amiga     }
{               number, please, tell us). A zero means to keep using the     }
{               same period used before.                                     }
{   Instrument: A number in range 0..63 meaning the number of the instrument }
{               which will be used for the note. A zero means use the same.  }
{   Command:    A number (no real range) of the way the note should be       }
{               played (i.e. Vibrato) a change in the playing sequence (i.e. }
{               pattern break) or a change in the general parameters of the  }
{               module player (i.e. set tempo). All the possible values are  }
{               defined in the TModCommand enumerated type below.            }
{   Parameter:  A parameter for the command. Its meaning differs from one    }
{               command to another. Sometimes each nibble is considered as a }
{               different parameter.                                         }
{____________________________________________________________________________}

TYPE tModCommand =
     (mcNone,       { 0 00 } { Just play the note, without any special option. }
      mcArpeggio,   { 0 xy } { Rotate through three notes rapidly. }
      mcTPortUp,    { 1 xx } { Tone Portamento Up:   Gradual change of tone towards high frequencies. }
      mcTPortDown,  { 2 xx } { Tone Portamento Down: Gradual change of tone towards low  frequencies. }
      mcNPortamento,{ 3 xy } { Note Portamento:      Gradual change of tone towards a given note.     }
      mcVibrato,    { 4 xy } { Vibrato: Frequency changes around the note. }
      mcT_VSlide,   { 5 xy } { Tone Port. Up + Volume slide: Parameter means vol. slide. }
      mcVib_VSlide, { 6 xy } { Vibrato       + Volume slide: Parameter means vol. slide. }
      mcTremolo,    { 7 xy } { Tremolo: I don't know for sure. Fast volume variations, I think. }
      mcNPI1,       { 8 xx } { Do Nothing (as far as I know). }
      mcSampleOffs, { 9 xx } { Start the sample from the middle. }
      mcVolSlide,   { A xy } { Volume slide: Gradual change in volume. }
      mcJumpPattern,{ B xx } { End pattern and continue from a different pattern sequence position. }
      mcSetVolume,  { C xx } { Set the volume of the sound. }
      mcEndPattern, { D xx } { Continue at the start of the next pattern. }
      mcExtended,   { E xy } { Extended set of commands (ProTracker). }
      mcSetTempo,   { F xx } { Set the tempo of the music, in 1/50ths of a second. }

      mcSetFilter,  { E 0x } { Set the output filter to the on or off value. }
      mcFinePortaUp,{ E 1x } { Like TPortUp,   but slower. }
      mcFinePortaDn,{ E 2x } { Like TPortDown, but slower. }
      mcGlissCtrl,  { E 3x } { ??? }
      mcVibCtrl,    { E 4x } { Set the vibrato waveform. }
      mcFineTune,   { E 5x } { Fine tune the frequency of the sound. }
      mcJumpLoop,   { E 6x } { Make a loop inside a pattern. }
      mcTremCtrl,   { E 7x } { Set the tremolo waveform (I think). }
      mcNPI2,       { E 8x } { Do Nothing (as far as I know). }
      mcRetrigNote, { E 9x } { ??? }
      mcVolFineUp,  { E Ax } { Like VolSlide, but slower and towards high frequencies. }
      mcVolFineDown,{ E Bx } { Like VolSlide, but slower and towards low  frequencies. }
      mcNoteCut,    { E Cx } { ??? }
      mcNoteDelay,  { E Dx } { Wait a little before starting note. }
      mcPattDelay,  { E Ex } { ??? }
      mcFunkIt,     { E Fx } { No idea, but sounds funny. }

      mcOktArp,     {      } { Oktalizer arpeggio  }
      mcOktArp2,    {      } { Oktalizer arpeggio2 }

      mcLast);

TYPE pNoCommandNote = ^tNoCommandNote;
     tNoCommandNote = RECORD
       Instrument : Byte;
       Period     : Word;
       Volume     : Byte;
     END;

     pCommandNote = ^tCommandNote;
     tCommandNote = RECORD
       Command    : tModCommand;
       Parameter  : Byte;
     END;

     pFullNote = ^tFullNote;
     tFullNote = RECORD
       CASE Byte OF
         0 : (Instrument : Byte;
              Period     : Word;
              Volume     : Byte;
              Command    : tModCommand;
              Parameter  : Byte);
         1 : (Note : tNoCommandNote;
              Comm : tCommandNote);
     END;

{----------------------------------------------------------------------------}
{ Definitions for handling the instruments used in the module.               }
{ Instruments are fragments of sampled sound (long arrays of Bytes which     }
{ describe the wave of the sound of the instrument). The samples used in     }
{ music modules have a default volume and also, they can have a loop (for    }
{ sustained instruments) and a fine tuning constant (not yet implemented).   }
{____________________________________________________________________________}

CONST MaxSample      = 65520;
      MaxInstruments = 255;

      LowQuality : Boolean = TRUE;

      { Properties }

      ipMonoFreq = $0001;  { Set if the instrument is played always at the same freq (not implemented). }
      ipLong     = $0002;  { Set if the instrument's sample is longer than 65520 Bytes.                 }

TYPE pSample = ^tSample;
     tSample = ARRAY[0..MaxSample-1] of ShortInt;

     tIProperties = Word; { Properties of the instrument. }

     pInstrumentRec = ^tInstrumentRec;
     tInstrumentRec = RECORD
       Len,                  { Length of the instrument's sampled image.                           }
       Reps,                 { Starting offset of the repeated portion.                            }
       Repl  : LongInt;      { Size of the repeated portion.                                       }
       Vol   : Byte;         { Default volume of the instrument (0..64)                            }
       Ftune : Byte;         { Fine tuning value for the instrument (not yet implemented).         }
       NAdj  : Word;         { Numerator of note adjutment.                                        }
       DAdj  : Word;         { Denominator of note adjutment.                                      }
       Data  : ^tSample;     { Pointer to the first  65520 Bytes of the sample.                    }
       Xtra  : ^tSample;     { Pointer to the second 65520 Bytes of the sample (if there is such). }
       Prop  : tIProperties; { Bit mapped properties value.                                        }
     END;

     pInstrument = ^tInstrument;
     tInstrument = OBJECT(TObject)
       Name  : pString;
       Instr : pInstrumentRec;

       CONSTRUCTOR Init;
       DESTRUCTOR  Done; VIRTUAL;

       PROCEDURE FreeContents;
       PROCEDURE Desample;

       PROCEDURE Change(Instrument : PInstrumentRec);
       FUNCTION  GetName : String;
       PROCEDURE SetName(S : String);
     END;

{----------------------------------------------------------------------------}
{ Definitions for handling the tracks of which patterns are built.           }
{ Tracks are lists of notes and command values of which the empty leading    }
{ and trailing blanks have been removed (obviated).                          }
{____________________________________________________________________________}

TYPE pNoteTrack = ^tNoteTrack;
     tNoteTrack = RECORD
       NoteOffset : Byte;
       NumNotes   : Byte;
       Notes      : ARRAY[0..255] of tNoCommandNote;
     END;

     pCommTrack = ^tCommTrack;
     tCommTrack = RECORD
       NoteOffset : Byte;
       NumNotes   : Byte;
       Notes      : ARRAY[0..255] of tCommandNote;
     END;

     pFullTrack = ^tFullTrack;
     tFullTrack = ARRAY[0..255] of tFullNote;

     pTrack = ^tTrack;
     tTrack = OBJECT(TObject)
       Name : pString;
       Note : pNoteTrack;
       Comm : pCommTrack;

       CONSTRUCTOR Init;
       DESTRUCTOR  Done; VIRTUAL;

       PROCEDURE FreeContents;

       PROCEDURE ChangeNote(At : Word; VAR FullNote : tFullNote);
       PROCEDURE GetNote(At : Word; VAR FullNote : tFullNote);

       PROCEDURE GetFullTrack(VAR Track : tFullTrack);
       PROCEDURE SetFullTrack(VAR Track : tFullTrack);

       FUNCTION GetName : String;
     END;

{----------------------------------------------------------------------------}
{ Definitions for handling the format of the patterns.                       }
{ Patterns are arrays of pointers to tracks (up to 12 tracks).               }
{ A music module can have up to 255 individual patterns, arranged in a       }
{ sequence of up to 255.                                                     }
{ Empty patterns are not counted.                                            }
{____________________________________________________________________________}

CONST MaxSequence     = 256;
      MaxPatterns     = 256;
      MaxPatternLines = 256;
      MaxChannels     = 16;

TYPE pPatternRec = ^tPatternRec;
     tPatternRec = RECORD
       NNotes   : Byte;
       NChans   : Byte;
       Tempo    : Byte;
       BPM      : Byte;
       Channels : ARRAY[1..MaxChannels] of Word;
     END;

     pPattern = ^tPattern;
     tPattern = OBJECT(TObject)
       Name : pString;
       Patt : pPatternRec;

       CONSTRUCTOR Init(Chans : Word);
       DESTRUCTOR  Done; VIRTUAL;

       PROCEDURE FreeContents;
       FUNCTION GetName : String;
     END;

     pPatternSequence = ^tPatternSequence;
     tPatternSequence = ARRAY[1..MaxSequence] of Byte;

{----------------------------------------------------------------------------}
{ General definitions for the song.                                          }
{____________________________________________________________________________}

TYPE pSongComment = ^tSongComment;
     tSongComment = ARRAY[1..16] of String[60];

IMPLEMENTATION

{----------------------------------------------------------------------------}
{ TInstrument object implementation.                                         }
{____________________________________________________________________________}

CONSTRUCTOR tInstrument.Init;
BEGIN
  tObject.Init;
END;


DESTRUCTOR tInstrument.Done;
BEGIN
  SetName('');
  FreeContents;
  tObject.Done;
END;

PROCEDURE tInstrument.FreeContents;
BEGIN
  If Instr = Nil then Exit;
  If Instr^.Len > 65520 then
  BEGIN
    FreeMem(Pointer(Instr^.Xtra), Instr^.Len - 65520);
    Instr^.Len := 65520;
  END;

  FreeMem(Pointer(Instr^.Data), Instr^.Len);
  FreeMem(Pointer(Instr), SizeOf(Instr^));
END;

PROCEDURE tInstrument.Change(Instrument : pInstrumentRec);
BEGIN
  FreeContents;
  If Instrument <> Nil then
  BEGIN
    GetMem(Pointer(Instr), SizeOf(Instr^));
    If Instr <> Nil then
    BEGIN
      Move(Instrument^, Instr^, SizeOf(Instr^));
      If Instr^.NAdj = 0 then
      BEGIN
        Instr^.NAdj := $2000;
        Instr^.DAdj := $2000;
      END;
      If LowQuality then DeSample;
    END;
  END;
END;

FUNCTION tInstrument.GetName : String;
BEGIN
  If Name <> Nil
    then GetName := Name^
    else GetName := '';
END;

PROCEDURE tInstrument.Desample;
VAR W        : Word;
    P        : Pointer;
    SizeFree : Word;
BEGIN
  WITH Instr^ DO
    If (Instr <> Nil) AND (Instr^.Data <> Nil) AND (Len > 128) AND ((Repl >= 2000) OR (Repl = 0)) then
    BEGIN
      For W := 0 to Len div 2 - 1 do
        Data^[W] := (Integer(Data^[W*2]) + Integer(Data^[W*2+1])) DIV 2;
      P := Ptr(Seg(Data^), Ofs(Data^) + Len DIV 2 + 7);
      P := Ptr(Seg(P^) + (Ofs(P^) SHR 4), Ofs(P^) AND $8);

      SizeFree := Len - (Word((Seg(p^) - Seg(Data^)) SHL 4) + Word(Ofs(P^) - Ofs(Data^)));

      FreeMem(P, SizeFree);

      Len  := Len  DIV 2;
      Reps := Reps DIV 2;
      Repl := Repl DIV 2;
      NAdj := NADJ  *  2;
    END;
END;

PROCEDURE tInstrument.SetName(S : String);
BEGIN
  If Name <> Nil then DisposeStr(Name);
  If S <> '' then Name := NewStr(S);
END;

{----------------------------------------------------------------------------}
{ TTrack object implementation.                                              }
{____________________________________________________________________________}

CONSTRUCTOR tTrack.Init;
BEGIN
  tObject.Init;
END;

DESTRUCTOR tTrack.Done;
BEGIN
  DisposeStr(Name);
  FreeContents;
  tObject.Done;
END;

PROCEDURE tTrack.FreeContents;
BEGIN
  If Note <> Nil then FreeMem(Pointer(Note), Note^.NumNotes*SizeOf(TNoCommandNote) + 2);
  If Comm <> Nil then FreeMem(Pointer(Comm), Comm^.NumNotes*SizeOf(TCommandNote)   + 2);
END;

PROCEDURE tTrack.ChangeNote(At : Word; VAR FullNote : tFullNote);
VAR Track : tFullTrack;
BEGIN
  GetFullTrack(Track);
  Track[At] := FullNote;
  SetFullTrack(Track);
END;

PROCEDURE tTrack.GetFullTrack(VAR Track : tFullTrack);
VAR I : Word;
BEGIN
  FillChar(Track, SizeOf(Track), 0);

  If Note <> Nil then
    For I := 0 to Note^.NumNotes do
      Track[I+Note^.NoteOffset].Note := Note^.Notes[I];

  If Comm <> Nil then
    For i := 0 to Note^.NumNotes do
      Track[I+Note^.NoteOffset].Note := Note^.Notes[I];
END;

PROCEDURE tTrack.SetFullTrack(VAR Track : tFullTrack);
VAR I     : Word;
    MNote : tNoteTrack;
    MComm : tCommTrack;
BEGIN
  FillChar(MNote, SizeOf(MNote), 0);
  FillChar(MComm, SizeOf(MComm), 0);
  For I := 0 to 255 do
  BEGIN
    If (Track[I].Instrument = 0) AND (Track[I].Period = 0) AND (Track[I].Volume = 0) then
    BEGIN
      If MNote.NoteOffset = I then Inc(MNote.NoteOffset);
    END else
    BEGIN
      MNote.NumNotes := I - MNote.NoteOffset + 1;
      MNote.Notes[I - MNote.NoteOffset] := Track[I].Note;
    END;

    If Track[I].Command = mcNone then
    BEGIN
      If MComm.NoteOffset = I then Inc(MComm.NoteOffset);
    END else
    BEGIN
      MComm.NumNotes := I - MComm.NoteOffset + 1;
      MComm.Notes[I - MComm.NoteOffset] := Track[I].Comm;
    END;
  END;

  FreeContents;

  GetMem(Pointer(Note), MNote.NumNotes*SizeOf(TNoCommandNote) + 2);
  GetMem(Pointer(Comm), MComm.NumNotes*SizeOf(TCommandNote)   + 2);

  If Note <> Nil then Move(MNote, Note^, MNote.NumNotes*SizeOf(TNoCommandNote) + 2);
  If Comm <> Nil then Move(MComm, Comm^, MComm.NumNotes*SizeOf(TCommandNote)   + 2);
END;

PROCEDURE tTrack.GetNote(At : Word; VAR FullNote : tFullNote);
BEGIN
  Dec(At);
  FillChar(FullNote, SizeOf(FullNote), 0);

  If (Note <> Nil) AND (At >= Note^.NoteOffset) AND (At < Note^.NoteOffset + Note^.NumNotes)
    then FullNote.Note := Note^.Notes[At - Note^.NoteOffset];

  If (Comm <> Nil) AND (At >= Comm^.NoteOffset) AND (At < Comm^.NoteOffset + Comm^.NumNotes)
    then FullNote.Comm := Comm^.Notes[At - Comm^.NoteOffset];
END;

FUNCTION tTrack.GetName : String;
BEGIN
  If Name <> Nil
    then GetName := Name^
    else GetName := '';
END;

{----------------------------------------------------------------------------}
{ TPattern object implementation.                                            }
{____________________________________________________________________________}

CONSTRUCTOR TPattern.Init(Chans : Word);
BEGIN
  tObject.Init;

  GetMem(Pointer(Patt), Chans*2 + 4);

  If Patt <> Nil then FillChar(Patt^, Chans*2 + 4, 0);
  Patt^.NChans := Chans;
END;

DESTRUCTOR TPattern.Done;
BEGIN
  DisposeStr(Name);
  FreeContents;
  tObject.Done;
END;

PROCEDURE TPattern.FreeContents;
BEGIN
  If Patt <> Nil then FreeMem(Pointer(Patt), Patt^.NChans*2 + 4);
END;

FUNCTION TPattern.GetName : String;
BEGIN
  If Name <> Nil
    then GetName := Name^
    else GetName := '';
END;

END.
