UNIT SongUnit;

INTERFACE
USES Dos, Objects, HexConversions, SongElements;

{----------------------------------------------------------------------------}
{ Song object definition.                                                    }
{____________________________________________________________________________}

TYPE tSongFileFormat =
     (mffUnknown,     { Unknown format O:-)                      }
      mffMod31M_K_,   { Protracker "M.K.".                       }
      mffMod31FLT4,   { Protracker "FLT4".                       }
      mffMod15,       { SoundTracker 15-instrument module.       }
      mffJMPlayer,    { JMPlayer module.                         }
      mffOktalizer,   { 8 voices Oktalizer MOD.           (.OKT) }
      mffComposer669, { 8 voices Composer-669.            (.669) }
      mffWow8,        { 8 voices Grave.                   (.WOW) }
      mffFastTracker, { 6 or 8 voices Triton FastTracker. (.MOD) }
      mffS3m,         { ScreamTracker 3.0                 (.S3M) }
      mffS2m,         { ScreamTracker 3.0 (beta)          (.S2M) }
      mffStm);        { ScreamTracker 2.x                 (.STM) }

     { Non fatal states }
     tSongStatus =
     (msNotLoaded,    { Not yet loaded                                         }
      msOK,           { Everything was Ok.                                     }
      msFileTooShort, { End of file premature (lot's of modules have this).    }

     { Fatal states }
      msFileOpenError,           { Could not open the .MOD file.                          }
      msOutOfMemory,             { There is not enough memory left. :-( Shouldn't happen. }
      msFileDamaged,             { Syntax checking error on module file.                  }
      msFileFormatNotSupported); { JMPlayer or ScreamTracker, for example.                }

TYPE pSong = ^tSong;
     tSong = OBJECT(tObject)
       { Desired data }
       SongStart : Word;
       SongLen   : Word;

       { General song data }
       Name         : pString;
       InsidePath   : pString;
       Comment      : pSongComment;
       FileDir      : pString;
       FileName     : NameStr;
       FileExt      : ExtStr;
       FirstTick    : Boolean;
       InitialTempo : Byte;
       InitialBPM   : Byte;
       Volume       : Byte;
       NumChannels  : Byte;

       { Instrument data }
       Instruments : tCollection;

       { Pattern sequence data }
       SequenceLength   : Word;
       SequenceRepStart : Word;
       PatternSequence  : pPatternSequence;
       PatternTempos    : pPatternSequence;

       Patterns         : tCollection;

       { Track data }
       Tracks : tCollection;

       { State data }
       Status      : tSongStatus;
       ErrorCode   : Word;
       ThereIsMore : Boolean;
       FileFormat  : tSongFileFormat;

       { Methods }
       CONSTRUCTOR Init;
       DESTRUCTOR  Done; VIRTUAL;

       PROCEDURE Load(VAR St : tStream);
       PROCEDURE Save(VAR St : tStream);

       PROCEDURE LoadFName(FName : PathStr);
       PROCEDURE SaveFName(FName : PathStr);

       PROCEDURE Free;
       PROCEDURE InitValues;
       PROCEDURE Empty;

       FUNCTION  GetErrorString : String;

       FUNCTION  GetName : String;
       FUNCTION  GetInsidePath : String;
       FUNCTION  GetInstrument(I : Word) : pInstrument;
       FUNCTION  GetTrack(I : Word) : pTrack;
       FUNCTION  GetPattern(I : Word) : pPattern;
       FUNCTION  GetPatternSeq(I : Word) : pPattern;
       FUNCTION  GetPatternSequence(Seq : Word) : Word;
       FUNCTION  GetPatternTempo(Seq : Word) : Word;
       PROCEDURE GetNote(Seq, Row, Chan : Word; VAR Note : tFullNote);

       PROCEDURE SetName(S : String);
       PROCEDURE SetInsidePath(S : String);
     END;

{----------------------------------------------------------------------------}
{ Header definition for the loaders.                                         }
{____________________________________________________________________________}

TYPE pSongHeader = ^tSongHeader;
     tSongHeader = ARRAY[0..2047] of Byte;

IMPLEMENTATION

USES SongUtils, UnkLoader, ModLoader, OktLoader, S3mLoader, StmLoader,
     Loader669, ExeLoader, Heaps, StrConst, AsciiZ, Filters;

{----------------------------------------------------------------------------}
{ Loaders definition.                                                        }
{____________________________________________________________________________}

TYPE tSongLoader = PROCEDURE(VAR Song : tSong; VAR St : tStream; VAR Header : tSongHeader);

CONST NumLoaders = 8;

      SongLoaders : ARRAY[1..NumLoaders] of tSongLoader =
      (LoadJMFileFormat,
       Load669FileFormat,
       LoadOktFileFormat,
       LoadS2mFileFormat,
       LoadS3mFileFormat,
       LoadStmFileFormat,
       LoadExeFileFormat,
       LoadModFileFormat);

{----------------------------------------------------------------------------}
{ TSong object.                                                              }
{____________________________________________________________________________}

CONSTRUCTOR tSong.Init;
BEGIN
  tObject.Init;
  InitValues;
END;

DESTRUCTOR tSong.Done;
BEGIN
  Free;
  tObject.Done;
END;

PROCEDURE tSong.Load(VAR St : tStream);
VAR Header : tSongHeader;
    I      : Word;
    Pos    : LongInt;
BEGIN
  Pos := St.GetPos;

  ThereIsMore := False;

  St.Read(Header, SizeOf(TSongHeader));

  If St.Status <> stOk then
  BEGIN
    Status    := msFileDamaged;
    ErrorCode := St.ErrorInfo;
    St.Done;
    Exit;
  END;

  I := 1;
  WHILE (I <= NumLoaders) AND (Status = msNotLoaded) DO
  BEGIN
    St.Seek(Pos);
    SongLoaders[I](PSong(@Self)^, St, Header);
    Inc(I);
  END;
END;

PROCEDURE tSong.LoadFName(FName : PathStr);
VAR St         : tDosStream;
    Dir        : DirStr;
    IPath      : String[12];
    OSongStart : Word;
    OSongLen   : Word;
BEGIN
  OSongStart := SongStart;
  OSongLen   := SongLen;
  IPath      := GetInsidePath;
  Empty;
  SetInsidePath(IPath);
  SongStart := OSongStart;
  SongLen   := OSongLen;

  FName := FExpand(FName);
  FSplit(FName, Dir, FileName, FileExt);
  FileDir := FullHeap.HNewStr(Dir);
  If FileExt = '' then FileExt := '.MOD';
  FName := Dir+FileName+FileExt;

  St.Init(FName, stOpenRead);

  If St.Status <> stOk then
  BEGIN
    Status    := msFileOpenError;
    ErrorCode := St.ErrorInfo;
    St.Done;
    Exit;
  END;

  Status    := msNotLoaded;
  ErrorCode := 0;

  Load(St);

  If Status <> msOk then ErrorCode := St.ErrorInfo;

  St.Done;
END;

PROCEDURE tSong.Save(VAR St : tStream);
BEGIN
END;

PROCEDURE tSong.SaveFName(FName : PathStr);
BEGIN
END;

FUNCTION tSong.GetErrorString : String;
BEGIN
  CASE Status OF
    msFileOpenError          : GetErrorString := GetString(StrFileOpenError);
    msOutOfMemory            : GetErrorString := GetString(StrOutOfMemory);
    msFileDamaged            : GetErrorString := GetString(StrFileDamaged);
    msFileTooShort           : GetErrorString := GetString(StrFileTooShort);
    msFileFormatNotSupported : GetErrorString := GetString(StrFileFormatNotSupported) +
                                                 GetString(StrFileFormats + Byte(FileFormat));
    else                       GetErrorString := '';
  END;
END;

FUNCTION tSong.GetName : String;
BEGIN
  If Name <> Nil
    then GetName := Name^
    else GetName := '';
END;

PROCEDURE tSong.SetName(S : String);
BEGIN
  If Name <> Nil then FullHeap.HDisposeStr(Name);
  If S <> '' then Name := FullHeap.HNewStr(S);
END;

FUNCTION tSong.GetInsidePath : String;
BEGIN
  If InsidePath <> Nil
    then GetInsidePath := InsidePath^
    else GetInsidePath := '';
END;

PROCEDURE tSong.SetInsidePath(S : String);
BEGIN
  If InsidePath <> Nil then FullHeap.HDisposeStr(InsidePath);
  If S <> '' then InsidePath := FullHeap.HNewStr(S);
END;

FUNCTION TSong.GetInstrument(I : Word) : pInstrument;
VAR Instrument : pInstrument;
    J          : Word;
LABEL Break;
BEGIN
  If I >= Instruments.Count then
  BEGIN
    For J := Instruments.Count to I do
    BEGIN
      Heap.HGetMem(Pointer(Instrument), SizeOf(TInstrument));
      If Instrument <> Nil then
      BEGIN
        Instrument^.Init;
        Instruments.AtInsert(J, Instrument);
      END else Goto Break;
    END;
Break:
    GetInstrument := Instrument;
  END else GetInstrument := PInstrument(Instruments.At(I));
END;

FUNCTION tSong.GetTrack(I : Word) : pTrack;
VAR Track : pTrack;
    J     : Word;
LABEL Break;
BEGIN
  If I >= Tracks.Count then
  BEGIN
    For J := Tracks.Count to I do
    BEGIN
      Heap.HGetMem(Pointer(Track), SizeOf(TTrack));
      If Track <> Nil then
      BEGIN
        Track^.Init;
        Tracks.AtInsert(J, Track);
      END else Goto Break;
    END;
Break:
    GetTrack := Track;
  END else GetTrack := PTrack(Tracks.At(I));
END;

FUNCTION tSong.GetPattern(I : Word) : pPattern;
VAR Pattern : pPattern;
    J       : Word;
LABEL Break;
BEGIN
  If I >= Patterns.Count then
  BEGIN
    For J := Patterns.Count to I do
    BEGIN
      Heap.HGetMem(Pointer(Pattern), SizeOf(TPattern));
      If Pattern <> Nil then
      BEGIN
        Pattern^.Init(NumChannels);
        Patterns.AtInsert(J, Pattern);
      END else Goto Break;
    END;
Break:
    GetPattern := Pattern;
  END else GetPattern := PPattern(Patterns.At(I));
END;

FUNCTION tSong.GetPatternSeq(I : Word) : pPattern;
BEGIN
  GetPatternSeq := GetPattern(GetPatternSequence(I));
END;

FUNCTION tSong.GetPatternSequence(Seq : Word) : Word;
BEGIN
  If PatternSequence <> Nil
    then GetPatternSequence := PatternSequence^[Word(Seq)]
    else GetPatternSequence := 0;
END;

FUNCTION TSong.GetPatternTempo(Seq : Word) : Word;
BEGIN
  If PatternTempos <> Nil
    then GetPatternTempo := PatternTempos^[Word(Seq)]
    else GetPatternTempo := 0;
END;

PROCEDURE TSong.GetNote(Seq, Row, Chan : Word; VAR Note : tFullNote);
VAR Patt  : pPattern;
    Track : pTrack;
    N     : Word;
    NOffs : Word;
BEGIN
  If PatternSequence <> Nil then
  BEGIN
    Patt := GetPatternSeq(Seq);
    If Patt <> Nil then
    BEGIN
      N     := Patt^.Patt^.Channels[Chan];
      Track := GetTrack(N);
      If Track <> Nil then
      BEGIN
        Track^.GetNote(Row, Note);
        Exit;
      END;
    END;
  END;

  FillChar(Note, SizeOf(Note), 0);
END;

PROCEDURE TSong.Free;
VAR I : Word;
BEGIN
  ASM CLI END;

  FullHeap.HDisposeStr(Name);
  FullHeap.HFreeMem   (Pointer(Comment), SizeOf(Comment^));
  FullHeap.HDisposeStr(FileDir);

  Instruments.Done;

  FullHeap.HFreeMem(Pointer(PatternSequence), SizeOf(PatternSequence^));
  FullHeap.HFreeMem(Pointer(PatternTempos),   SizeOf(PatternTempos^));
  Patterns.Done;

  Tracks.Done;

  ASM STI END;
END;

PROCEDURE TSong.InitValues;
BEGIN
  SongStart    := 1;
  SongLen      := MaxSequence;

  Name         := Nil;
  InsidePath   := Nil;
  Comment      := Nil;
  FileDir      := Nil;
  FileName     := '';
  FileExt      := '';
  FirstTick    := False;
  InitialTempo := 1;
  InitialBPM   := 1;
  Volume       := 0;
  NumChannels  := 0;

  Instruments.Init(32, 32);

  SequenceLength   := 0;
  SequenceRepStart := 0;

  FullHeap.HGetMem(Pointer(PatternSequence), SizeOf(PatternSequence^));
  FullHeap.HGetMem(Pointer(PatternTempos),   SizeOf(PatternTempos^));
  If PatternSequence <> Nil then FillChar(PatternSequence^, SizeOf(PatternSequence^), 0);
  If PatternTempos <> Nil then FillChar(PatternTempos^, SizeOf(PatternTempos^), 0);
  Patterns.Init(64, 64);

  Tracks.Init(256, 256);

  Status      := msNotLoaded;
  ErrorCode   := 0;
  ThereIsMore := False;
  FileFormat  := mffUnknown;
END;

PROCEDURE TSong.Empty;
BEGIN
  Free;
  InitValues;
END;

END.
