UNIT SoundDevices;

INTERFACE
USES SongElements;

{----------------------------------------------------------------------------}
{ Device configuration definitions.                                          }
{____________________________________________________________________________}

TYPE tDevName = String[50]; { Name/description of device.                                    }
     tDevID   = String[20]; { Device identification string.                                  }

     tProc            = PROCEDURE;                  { Generic procedure without parameters.                }
     tNameProc        = FUNCTION : tDevName;        { Procedure that returns the name of a device.         }
     tInitDevProc     = PROCEDURE(Hz : Word);       { Device initialisation procedure.                     }
     tChgHzProc       = PROCEDURE(Hz : Word);       { Sample rate change procedure.                        }
     tGetRealFreqProc = FUNCTION(Hz : Word) : Word; { Returns the real sampling freq. when Hz is selected. }
     tDetectProc      = FUNCTION : Boolean;         { Device autodetection procedure.                      }

{ Device record for including in a linked list.                            }
TYPE pSoundDevice = ^tSoundDevice;
     tSoundDevice = RECORD
       DevID : tDevID;                     { Device ID string.                                                        }
       DMA   : Boolean;                    { True if the device uses DMA output (shouldn't be needed).                }

       Name            : tNameProc;        { Device name.                                                       }
       Autodetect      : tDetectProc;      { Autodetection procedure.                                           }
       InitRut         : tInitDevProc;     { Initialisation procedure.                                          }
       ChgHzProc       : tChgHzProc;       { Sample rate change procedure.                                      }
       GetRealFreqProc : tGetRealFreqProc; { Real sampling freq.                                                }
       TimerHandler,                       { INT 8 handler for the device.                                      }
       PollRut         : tProc;            { Routine to be executed for active polling (hand made).             }
       EndRut          : tProc;            { Device closing procedure.                                          }

       Next : pSoundDevice;                { Next record in the list.                                             }
     END;

CONST NumDevices   : Byte         = 0;   { Count of the number of installed devices.    }
      ActiveDevice : pSoundDevice = Nil; { Device being used right now.                 }

{----------------------------------------------------------------------------}
{ Device Stack.                                                              }
{____________________________________________________________________________}

CONST DevStkSize = 500;
VAR DevStack : ARRAY[1..DevStkSize] of Byte;
    DevSS     : Word;
    DevSP     : Word;

{----------------------------------------------------------------------------}
{ Sample buffers definition.                                                 }
{____________________________________________________________________________}

TYPE tDataType = (dtShortInt, dtInteger);        { Data type of the samples. }

     tIntBuff   = ARRAY[0..32760] of Integer;    { Data types for big arrays. }
     tShortBuff = ARRAY[0..65520] of ShortInt;

     pIntBuff   = ^tIntBuff;                     { Idem. }
     pShortBuff = ^tShortBuff;

     pSampleBuffer = ^tSampleBuffer;   { PCM Buffer. }
     tSampleBuffer = RECORD
       InUse    : Boolean;             { True while it's being used by the device.       }
       NSamples,                       { Size of the buffer in samples.                  }
       RateHz   : Word;                { Sampling frequency.                             }
       Channels : Byte;                { 1 or 4, channels contained in the buffer.       }
       CASE DataType : tDataType OF
         dtInteger  : (IData : PIntBuff);   { Pointer to the buffer.                    }
         dtShortInt : (SData : PShortBuff);
     END;

CONST MaxChannels  = SongElements.MaxChannels;
      Sounding     : Pointer = Nil;  { Buffer that is actually sounding (NON-DMA only). }
      SoundLeft    : Word    = 0;    { Number of samples left in the buffer.            }
      NumChannels  : Byte    = 1;    { Number of channels in the buffer.                }
      ChannelIncr  : Word    = 1;    { Size of one sample in the buffer.                }

{----------------------------------------------------------------------------}
{ DMA buffers definition.                                                    }
{____________________________________________________________________________}

CONST DMABufferSize = 5*700;     { Size of the buffers. }

VAR DMABufferPtr : Pointer;                         { DMA-fixed ;-) pointers for   }
    DMABuffer    : Pointer;                         { buffers 1 and 2.             }
    DMABufferEnd : Word;

{----------------------------------------------------------------------------}
{ Hardware parameters.                                                       }
{____________________________________________________________________________}

CONST DefaultHz                   = 16000;     { Default sampling rate.                                }
      DeviceIdling      : Boolean = True;      { True if there are no samples sounding.                }
      TimerHz           : Word    = DefaultHz; { Clock frequency of the INT 8 timer.                   }
      LastHz            : Word    = 0;         { Older INT 8 frequency (for detecting change).         }
      SoundHz           : Word    = DefaultHz; { Sampling frequency of the sound.                      }
      DesiredHz         : Word    = DefaultHz; { Desired sampling frequency of the sound.              }
      SystemClockCount  : Word    = 0;         { Clock count for calling the original INT 8.           }
      SystemClockIncr   : Word    = 0;         { Increment for calling the original INT 8.             }
      TimerVal          : Word    = 0;         { Value given to the INT 8 timer.                       }
      DeviceInitialized : Boolean = False;     { True if a device has already been initialized.        }
      DMAOffset         : Word    = 1;         { Number of samples to discard in DMA transferences.    }
      HzChanged         : Boolean = False;
      DoEqualice        : Boolean = False;

{----------------------------------------------------------------------------}
{ Periodic process.                                                          }
{____________________________________________________________________________}

VAR PeriodicProc : tProc;    { Periodic process (normally a music interpreter).           }

CONST PeriodicHz    : Byte = 0; { Frequency for calling the periodic process.                }
      PeriodicStart : Word = 1; { Countdown starting point (NON-DMA only).                   }
      PeriodicCount : Word = 0; { Countdown.               (idem).                           }

{----------------------------------------------------------------------------}
{ Buffer provider definitions.                                               }
{____________________________________________________________________________}

TYPE tAskBufferProc = FUNCTION : pSampleBuffer; { Buffer provider function. }

VAR AskBufferProc  : tAskBufferProc; { Pointer to the buffer provider.             }
    ActualBuffer,                    { Buffer being used.                          }
    NextBuffer     : pSampleBuffer;  { Buffer that will be used next.              }
    PleaseFallback : Word;           { Set >0 if there are no buffers available.   }

{----------------------------------------------------------------------------}
{ Functions to be used by devices only.                                      }
{____________________________________________________________________________}

FUNCTION  InitDevice(Device : pSoundDevice) : Word; { Used to declare a device.                            }
PROCEDURE PollDevice;                               { Used to manually poll the device, if it is required. }
PROCEDURE CalcTimerData(Hz : Word);                 { Used to calculate the different Hz variables.        }
PROCEDURE DefaultChgHz(Hz : Word);                  { Used as a default TChgHzProc.                        }
FUNCTION  GetRealFreq(Hz : Word) : Word;            { Used as a default TRealFreqProc.                     }
PROCEDURE InitTimer;                                { Used to reinitialise the timer after a freq. change. }
FUNCTION  DoGetBuffer : Word;                       { Used to get the next buffer prepared.                }
PROCEDURE CSOldInt8;                                { Used to call the old INT 8 vector.                   }

{----------------------------------------------------------------------------}
{ Functions to be used by the sound generators only.                         }
{____________________________________________________________________________}

PROCEDURE SetDevice(P : pSoundDevice);                     { Used to initialise a buffer for output.  }
FUNCTION  IndexDevice(I : Word) : pSoundDevice;            { Used to index the devices.               }
FUNCTION  LocateDevice(ID : String) : pSoundDevice;        { Used to find a given device.             }
PROCEDURE SetPeriodicProc(Proc : tProc; PerSecond : Word); { Used to initialise the periodic process. }
PROCEDURE SetBufferAsker(Proc : tAskBufferProc);           { Used to initialise the buffer asker.     }
PROCEDURE StartSampling;                                   { Used to start the sound output.          }
PROCEDURE EndSampling;                                     { Used to end the sound output.            }

PROCEDURE InitSoundDevices;

IMPLEMENTATION
USES Dos;

{----------------------------------------------------------------------------}
{ Internal data.                                                             }
{____________________________________________________________________________}

CONST DeviceList       : PSoundDevice = Nil;   { Linked list of all devices.                     }
      OldInt8          : Pointer      = Nil;   { Pointer to the original INT 8.                  }
      IntInstalled     : Boolean      = False; { True if the INT 8 handler is already installed. }

{----------------------------------------------------------------------------}
{ Null procedures used in the unit.                                          }
{____________________________________________________________________________}

PROCEDURE NullProcedure; FAR; ASSEMBLER;
ASM
END;

FUNCTION  NullBufferProc : PSampleBuffer; FAR;
BEGIN
  NullBufferProc := Nil;
END;

PROCEDURE NullInt; ASSEMBLER;
ASM
  PUSH    AX
  MOV     AL,$20
  OUT     $20,AL
  POP     AX
  IRET
END;

{----------------------------------------------------------------------------}
{ A little bit messy, but it implements an easy jump to the original INT 8.  }
{____________________________________________________________________________}

PROCEDURE CSOldInt8; ASSEMBLER;
ASM
  JMP FAR PTR CSOldInt8;
END;

TYPE pCSOldInt8 = ^tCSOldInt8;
     tCSOldInt8 = RECORD
       JMP : Byte;
       Int : Pointer;
     END;

VAR _CSOldInt8 : pCSOldInt8;

{----------------------------------------------------------------------------}
{ Periodic process implementation.                                           }
{____________________________________________________________________________}

PROCEDURE InitPeriodic;
BEGIN
  If PeriodicHz = 0 then
  BEGIN
    PeriodicStart   := 0;
    PeriodicCount   := 0;
    SystemClockIncr := TimerVal;
  END else
  BEGIN
    PeriodicStart := TimerHz DIV PeriodicHz;
    If PeriodicStart = 0 then PeriodicStart := 1;
    PeriodicCount := 1;
    SystemClockIncr := TimerVal * PeriodicStart;
  END;
END;

PROCEDURE SetPeriodicProc(Proc : tProc; PerSecond : Word);
BEGIN
  ASM
    PUSHF
    CLI
    LES     BX,[Proc]
    MOV     Word PTR [PeriodicProc],BX;
    MOV     Word PTR [PeriodicProc+2],ES;
    POPF
  END;
  PeriodicHz   := PerSecond;
  InitPeriodic;
END;

{----------------------------------------------------------------------------}
{ Hardware and interrupt handling procedures.                                }
{____________________________________________________________________________}

PROCEDURE OriginalHwTimer; ASSEMBLER;
ASM
  MOV     AL,54       { Select timer 0, sequential access and continuous mode. }
  OUT     43h,AL
  XOR     AL,AL       { Set the counter to 0 (65536). }
  OUT     40h,AL      { Lower byte of the counter.    }
  OUT     40h,AL      { Higher byte.                  }
END;

PROCEDURE SetHwTimer(Value : Word); ASSEMBLER;
ASM
  MOV     AL,54       { Select timer 0, sequential access and continuous mode. }
  OUT     43h,AL
  MOV     AX,value
  OUT     40h,AL      { Lower byte of the counter.    }
  XCHG    Ah,AL
  OUT     40h,AL      { Higher byte.                  }
END;

PROCEDURE RestoreTimer;
BEGIN
  If IntInstalled then
  BEGIN
    SetIntVec(8, OldInt8);
    OriginalHwTimer;
    IntInstalled := False;
  END;
END;

PROCEDURE InitTimer;
BEGIN
  InitPeriodic;

  If NOT IntInstalled then
  BEGIN
    IntInstalled := True;
    GetIntVec(8, OldInt8);
    _CSOldInt8^.Int := OldInt8;
    SetIntVec(8, @ActiveDevice^.TimerHandler);
  END;

  SetHwTimer(TimerVal);
END;

{----------------------------------------------------------------------------}
{ Procedures exported for the sound generator.                               }
{____________________________________________________________________________}

PROCEDURE StartSampling;
BEGIN
  If NOT DeviceInitialized then RestoreTimer;

  ActualBuffer      := Nil;
  NextBuffer        := Nil;
  SoundLeft         := 0;
  PleaseFallBack    := 0;
  DeviceIdling      := True;
  DMABufferPtr      := DMABuffer;

  FillChar(DMABuffer^, DMABufferSize, $80);

  If (ActiveDevice <> Nil) then
  BEGIN
    ASM CLI END;

    DeviceInitialized := True;
    ActiveDevice^.InitRut(DesiredHz);

    ASM STI END;
  END;
END;

PROCEDURE EndSampling;
BEGIN
  If (ActiveDevice <> Nil) AND DeviceInitialized then
  BEGIN
    ASM CLI END;
    ActiveDevice^.EndRut;
    RestoreTimer;
    ASM STI END;
    DeviceInitialized := False;
  END;
END;

PROCEDURE SetBufferAsker(Proc : tAskBufferProc);
BEGIN
  ASM CLI END;
  AskBufferProc := Proc;
  ASM STI END;
END;

PROCEDURE SetDevice(P : pSoundDevice);
BEGIN
  If p <> Nil then
  BEGIN
    If DeviceInitialized then
    BEGIN
      EndSampling;
      ActiveDevice := P;
      StartSampling;
    END else ActiveDevice := P;
  END;
END;

FUNCTION LocateDevice(ID : String) : pSoundDevice;

  FUNCTION NotInStr(VAR S, SS : String) : Boolean;
  VAR I : Word;
  BEGIN
    NotInStr := True;
    If Length(SS) > Length(S) then Exit;
    For I := 1 to Length(SS) DO
      If UpCase(S[I]) <> UpCase(SS[I]) then Exit;
    NotInStr := False;
  END;

VAR P : pSoundDevice;
BEGIN
  P := DeviceList;
  WHILE (P <> Nil) AND NotInStr(P^.DevID, ID) DO
    P := P^.Next;
  LocateDevice := P;
END;

FUNCTION IndexDevice(I : Word) : pSoundDevice;
VAR P : pSoundDevice;
BEGIN
  P := DeviceList;
  Dec(I);
  WHILE (P <> Nil) AND (I > 0) DO
  BEGIN
    P := P^.Next;
    Dec(I);
  END;

  IndexDevice := P;
END;

{----------------------------------------------------------------------------}
{ Implementation of some procedures exported to the device controllers.       }
{____________________________________________________________________________}

FUNCTION InitDevice(Device : pSoundDevice) : Word;
BEGIN
  Device^.Next := DeviceList;
  DeviceList   := Device;
  If ActiveDevice = Nil then SetDevice(Device);
  Inc(NumDevices);
END;

PROCEDURE PollDevice;
BEGIN
  ActiveDevice^.PollRut;
END;

FUNCTION GetRealFreq(Hz : Word) : Word;
VAR I    : Word;
    NHz1 : Word;
    NHz2 : Word;
BEGIN
  If Hz = 0 then Hz := 1;
  I := 1193180 DIV Hz;

  NHz1 := 1193180 DIV  i;
  NHz2 := 1193180 DIV (i + 1);
  If Abs(Integer(NHz1 - Hz)) > Abs(Integer(NHz2 - Hz)) then NHz1 := NHz2;

  GetRealFreq := NHz1;
END;

PROCEDURE CalcTimerData(Hz : Word);
BEGIN
  Hz := GetRealFreq(Hz);

  If Hz = 0 then TimerVal := $FFFF else TimerVal := 1193180 DIV Hz;

  TimerHz := 1193180 DIV TimerVal;
  SoundHz := TimerHz;
END;

PROCEDURE DefaultChgHz(Hz : Word);
BEGIN
    CalcTimerData(Hz);
    InitTimer;
  END;


FUNCTION DoGetBuffer  : Word;
CONST Semaphore : Byte    = 0;
      Size      : Word    = 1;
BEGIN
  DoGetBuffer := 0;
  If Semaphore > 0 then Exit;
  Inc(Semaphore);

  If ActualBuffer <> Nil then
  BEGIN
    Size := ActualBuffer^.NSamples;
    ActualBuffer^.InUse := False;
  END;

  ActualBuffer := NextBuffer;
  NextBuffer   := AskBufferProc;

  If ActualBuffer = Nil then
  BEGIN
    ActualBuffer := NextBuffer;
    If ActualBuffer <> Nil then
    BEGIN
      ActualBuffer^.InUse := True;
      NextBuffer          := AskBufferProc;
    END;
  END;

  If NextBuffer <> Nil then NextBuffer^.InUse := True;

  If ActualBuffer = Nil then
  BEGIN
    If (Size <> 1) AND (NOT ActiveDevice^.DMA) then Inc(PleaseFallBack);
    SoundLeft      := 0;
    If NOT ActiveDevice^.DMA then
    BEGIN
      PeriodicCount := 1;
      LastHz        := PeriodicHz;
      ActiveDevice^.ChgHzProc(LastHz);
    END;
  END else
  BEGIN
    If (LastHz <> ActualBuffer^.RateHz) then
    BEGIN
      LastHz := ActualBuffer^.RateHz;
      ActiveDevice^.ChgHzProc(LastHz);

      HzChanged := True;
    END;

    Sounding     := ActualBuffer^.IData;
    SoundLeft    := ActualBuffer^.NSamples;
    NumChannels  := ActualBuffer^.Channels;
    ChannelIncr  := ActualBuffer^.Channels * (ORD(ActualBuffer^.DataType)+1);

    If ActiveDevice^.DMA then
    BEGIN
      If SoundLeft > DMAOffset + 5
        then Dec(SoundLeft, DMAOffset)
        else SoundLeft := 5;
    END;
  END;

  DoGetBuffer := SoundLeft;
{  WriteNum(40, SoundLeft, $70);}

  Dec(Semaphore);
END;

{----------------------------------------------------------------------------}
{ Unit initialisation.                                                       }
{____________________________________________________________________________}

PROCEDURE InitSoundDevices;
TYPE pFreeBlock = ^tFreeBlock;
     tFreeBlock = RECORD
       Next : pFreeBlock;
       Size : Pointer;
     END;

VAR L        : LongInt;
    PtrFree  : Pointer;
    OldHPtr  : Pointer;
    p        : pFreeBlock;
    OffsFree : Word;

BEGIN
  _CSOldInt8 := @CSOldInt8;

  PeriodicProc  := NullProcedure;
  AskBufferProc := NullBufferProc;

  DMABuffer := HeapPtr;

  L := (LongInt(Seg(DMABuffer^)) SHL 4) + OFS(DMABuffer^);
  PtrFree  := HeapPtr;
  OffsFree := 0;

  If LongInt(Word(l)) + DMABufferSize > 65536 then
  BEGIN
    OffsFree := 65536 - LongInt(Word(L));
    L := (L AND $F0000) + $10000;
  END;

  DMABuffer    := Ptr((L SHR 4) AND $F000, Word(L));
  DMABufferPtr := DMABuffer;
  DMABufferEnd := OFS(DMABuffer^) + DMABufferSize;

  OldHPtr      := HeapPtr;
  HeapPtr      := Ptr((L + DMABufferSize + 16) SHR 4, 0);
  If OldHPtr = FreeList then FreeList := HeapPtr else
  BEGIN
    P := FreeList;
    WHILE P^.Next <> OldHPtr DO
      P := P^.Next;
      P^.Next := HeapPtr;
    END;

  FillChar(HeapPtr^, 8, 0);

  If OffsFree > 0 then FreeMem(PtrFree, OffsFree);
END;

END.

